package net.sf.itcb.common.portlet.vaadin;

import java.net.SocketException;

import javax.portlet.ActionRequest;
import javax.portlet.ActionResponse;
import javax.portlet.EventRequest;
import javax.portlet.EventResponse;
import javax.portlet.PortletMode;
import javax.portlet.PortletModeException;
import javax.portlet.PortletURL;
import javax.portlet.RenderRequest;
import javax.portlet.RenderResponse;
import javax.portlet.ResourceRequest;
import javax.portlet.ResourceResponse;
import javax.portlet.WindowState;
import javax.portlet.WindowStateException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import net.sf.itcb.common.business.core.ItcbApplicationContextHolder;
import net.sf.itcb.common.portlet.vaadin.page.PageMappingProcessor;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.liferay.portal.kernel.management.jmx.GetAttributesAction;
import com.vaadin.Application;
import com.vaadin.terminal.ErrorMessage;
import com.vaadin.terminal.ParameterHandler;
import com.vaadin.terminal.SystemError;
import com.vaadin.terminal.Terminal;
import com.vaadin.terminal.URIHandler;
import com.vaadin.terminal.VariableOwner;
import com.vaadin.terminal.gwt.server.ChangeVariablesErrorEvent;
import com.vaadin.terminal.gwt.server.HttpServletRequestListener;
import com.vaadin.terminal.gwt.server.PortletApplicationContext2;
import com.vaadin.terminal.gwt.server.PortletApplicationContext2.PortletListener;
import com.vaadin.terminal.gwt.server.WebApplicationContext;
import com.vaadin.ui.AbstractComponent;
import com.vaadin.ui.Button;
import com.vaadin.ui.Panel;
import com.vaadin.ui.TabSheet;
import com.vaadin.ui.TabSheet.SelectedTabChangeEvent;
import com.vaadin.ui.TabSheet.SelectedTabChangeListener;
import com.vaadin.ui.TabSheet.Tab;
import com.vaadin.ui.VerticalLayout;
import com.vaadin.ui.Window;
import com.vaadin.ui.Window.CloseEvent;



/**
 * this abstract class is the base vaadin application portlet used by all the vaadin application
 * It manages common treatments :
 * <ul>
 * <li>user storage in a spring Context {@link org.springframework.security.core.context.SecurityContextHolder}</li>
 * <li>local storage in a spring Context {@link org.springframework.context.i18n.LocaleContextHolder}</li> 
 * <li>user storage in MDC context in order to be logged {@link org.slf4j.MDC}</li>
 * <li>create the application and all the views (VIEW, EDIT, HELP) when the application is run inside a portal or directly in an application server as a servlet</li>
 * <li>delegates the views contents to child classes</li>
 * </ul>
 * 
 * TODO (Feature request  3155816): Request in an portal free mode (running in Tomcat as servlet) is ok but 2 things has to be done :
 * <ul>
 * <li>Request are now RenderRequest but it has to be also possible to use HttpServletRequest</li>
 * <li>Users have to be mocked when we are in Servlet mode</li>
 * <li>Edit View and Help mode have to be displayed with a tabs screen instead of the 3 actual buttons.</li>
 * </ul>
 * In child class and all the other classes : Services, DAO... The user will be available by calling the spring context : SecurityContextHolder.getContext().getAuthentication().
 * 
 * @author Pierre Le Roux
 */
public abstract class AbstractItcbPortletApplication extends Application implements HttpServletRequestListener, PortletListener {
	
	private static final long serialVersionUID = 1L;
	
	protected final Logger log = LoggerFactory.getLogger(getClass());
	
	protected Window mainWindow;
	
	protected String preferencesURL;
	
	protected PageMappingProcessor viewModePageMappingProcessor;
	
	protected PageMappingProcessor editModePageMappingProcessor;
	
	protected PageMappingProcessor helpModePageMappingProcessor;
	
	protected PageMappingProcessor currentPageMappingProcessor;
	
	protected boolean hasToCloseApplicationOnExit = false;
	
	private HttpServletRequest servletRequest = null;
	
	public void init() {
		mainWindow = new Window(getWindowName());
		setMainWindow(mainWindow);
		
		// The portlet is launched in a portal
		if (getContext() instanceof PortletApplicationContext2) {
			PortletApplicationContext2 ctx =
				(PortletApplicationContext2) getContext();		
			log.debug("init" + getUser());
			ctx.addPortletListener(this, this);
		}
		// The portlet is launched as a webapplication in a web app
		else if (getContext() instanceof WebApplicationContext) {
			WebApplicationContext ctx =
				(WebApplicationContext) getContext();
			manageServletContext(ctx);
			
				// TODO Add mocked request and response with spring configuration instead of null values
				// For now, portlet is ok outside a portal if the portlet doesn't use request and response objects
				//getMainWindow().addComponent(new PortletDevView(getPortletModeView(null, null), getPortletModeEdit(null, null), getPortletModeHelp(null,null)));

			
		}

	}
	
	@Override
	public void terminalError(Terminal.ErrorEvent event) {
		final Throwable t = event.getThrowable();
        if (t instanceof SocketException) {
            // Most likely client browser closed socket
            return;
        }

        // Finds the original source of the error/exception
        Object owner = null;
        if (event instanceof VariableOwner.ErrorEvent) {
            owner = ((VariableOwner.ErrorEvent) event).getVariableOwner();
        } else if (event instanceof URIHandler.ErrorEvent) {
            owner = ((URIHandler.ErrorEvent) event).getURIHandler();
        } else if (event instanceof ParameterHandler.ErrorEvent) {
            owner = ((ParameterHandler.ErrorEvent) event).getParameterHandler();
        } else if (event instanceof ChangeVariablesErrorEvent) {
            owner = ((ChangeVariablesErrorEvent) event).getComponent();
        }

        // Shows the error in AbstractComponent
        if (owner instanceof AbstractComponent && !(owner instanceof Button) ) { // We don't want to display error on Button. Only on components
            if (t instanceof ErrorMessage) {
                ((AbstractComponent) owner).setComponentError((ErrorMessage) t);
            } else {
                ((AbstractComponent) owner)
                        .setComponentError(new SystemError(t));
            }
        }

        // also print the error on console
        log.error("Terminal error:", t);
		
		try  {
			currentPageMappingProcessor.handleError(event);
		}
		catch(Exception errorInExceptionManagementException) {
			errorInExceptionManagementException.printStackTrace();
		}
	}

	@Override
	public void onRequestStart(HttpServletRequest request,
			HttpServletResponse response) {	
		this.servletRequest = request;
		
	}
	
	@Override
	public void onRequestEnd(HttpServletRequest request,
			HttpServletResponse response) {
		// Nothing
		
	}

	/**
	 * Manage a servletContext mode
	 * TODO Not totally available for now
	 */
	private void manageServletContext(WebApplicationContext ctx) {
		mainWindow.removeAllComponents();
		TabSheet modeTabs = new TabSheet();
		modeTabs.setSizeFull();
		
		final VerticalLayout viewPanel = new VerticalLayout();
		viewPanel.setSizeFull();
		final VerticalLayout editPanel = new VerticalLayout();
		editPanel.setSizeFull();
		final VerticalLayout helpPanel = new VerticalLayout();
		helpPanel.setSizeFull();
		modeTabs.addTab(viewPanel, PortletMode.VIEW.toString(), null);
		modeTabs.addTab(editPanel, PortletMode.EDIT.toString(), null);
		modeTabs.addTab(helpPanel, PortletMode.HELP.toString(), null);
		
		modeTabs.setSelectedTab(viewPanel);
		currentPageMappingProcessor = viewModePageMappingProcessor;
		modeTabs.addListener(new SelectedTabChangeListener() {

			private static final long serialVersionUID = 1L;

			@Override
			public void selectedTabChange(SelectedTabChangeEvent event) {
				TabSheet tabsheet = event.getTabSheet();
		        Tab tab = tabsheet.getTab(tabsheet.getSelectedTab());
			
				if (PortletMode.VIEW.equals(tab.getCaption())) {
					currentPageMappingProcessor = viewModePageMappingProcessor;
					//currentPageMappingProcessor.setPanel(viewPanel); // FIXME Change it to make it more modular with PanelSpecific stuffs
				}
				else if (PortletMode.EDIT.equals(tab.getCaption())) {
					currentPageMappingProcessor = editModePageMappingProcessor;
					//currentPageMappingProcessor.setPanel(editPanel); // FIXME Change it to make it more modular with PanelSpecific stuffs
				}
				else if (PortletMode.HELP.equals(tab.getCaption())) {
					currentPageMappingProcessor = helpModePageMappingProcessor;
					//currentPageMappingProcessor.setPanel(helpPanel); // FIXME Change it to make it more modular with PanelSpecific stuffs
				}
				
				currentPageMappingProcessor.setRequest(servletRequest);
				currentPageMappingProcessor.setApplication(AbstractItcbPortletApplication.this);
				if(currentPageMappingProcessor.isAutomaticDisplay()) {
					currentPageMappingProcessor.displayDefaultPage();
				}
				
			}
			
		});
		
		
		mainWindow.addComponent(modeTabs);		
		
	}

	public void setHasToCloseApplicationOnExit(boolean hasToCloseApplicationOnExit) {
		this.hasToCloseApplicationOnExit = hasToCloseApplicationOnExit;
	}

	public PageMappingProcessor getCurrentPageMappingProcessor() {
		return currentPageMappingProcessor;
	}
	

	protected String getWindowName() {
		return "";
	}
	
	
	/**
	 * Retrieves the componentContainer in which the content will be displayed
	 * If the PortletApplication implements {@link PanelSpecificPortlet}, 
	 * it will display content in the specific panel depending on the portletMode
	 * @param portletMode the portlet mode
	 * @return the panel in which the content will be written
	 */
	protected Panel getComponentContainer(PortletMode portletMode) {
		if(this instanceof PanelSpecificPortlet) {
			if(PortletMode.VIEW.equals(portletMode)) {
				return ((PanelSpecificPortlet)this).getPanelView();
			}
			else if(PortletMode.EDIT.equals(portletMode)) {
				return ((PanelSpecificPortlet)this).getPanelEdit();
			}
			else if(PortletMode.HELP.equals(portletMode)) {
				return ((PanelSpecificPortlet)this).getPanelHelp();
			}
		}
		return getMainWindow();
	}
	
	

	public void handleActionRequest(
		ActionRequest request, ActionResponse response, Window window) {
	}

	public void handleRenderRequest(
		RenderRequest request, RenderResponse response, Window window) {
		viewModePageMappingProcessor = ItcbApplicationContextHolder.getContext().getBean("viewModePageMappingProcessor", PageMappingProcessor.class);
		editModePageMappingProcessor = ItcbApplicationContextHolder.getContext().getBean("editModePageMappingProcessor", PageMappingProcessor.class);
		helpModePageMappingProcessor = ItcbApplicationContextHolder.getContext().getBean("helpModePageMappingProcessor", PageMappingProcessor.class);
		
		mainWindow = new Window(getWindowName());
		setMainWindow(mainWindow);
		
		mainWindow.addListener(new Window.CloseListener() {
			private static final long serialVersionUID = 1L;

			@Override
		    public void windowClose(CloseEvent poEvent)
			{
				log.debug("Closing Application window");
				
				// Close the application or not when the window is closed
				if(hasToCloseApplicationOnExit) {
					mainWindow = null;
					AbstractItcbPortletApplication.this.close();
				}
		    }
		});
		
		if (PortletMode.VIEW.equals(request.getPortletMode())) {
			currentPageMappingProcessor = viewModePageMappingProcessor;
		}
		else if (PortletMode.EDIT.equals(request.getPortletMode())) {
			currentPageMappingProcessor = editModePageMappingProcessor;
		}
		else if (PortletMode.HELP.equals(request.getPortletMode())) {
			currentPageMappingProcessor = helpModePageMappingProcessor;
		}
		if(currentPageMappingProcessor == null) {
			log.error("the "+ request.getPortletMode() +" associated mapping variable has to be set in spring context. The bean name has to be : " + request.getPortletMode() + "ModePageMappingProcessor");
			return;
		}
		currentPageMappingProcessor.setRequest(request);
		currentPageMappingProcessor.setPanel(getComponentContainer(request.getPortletMode()));
		currentPageMappingProcessor.setApplication(AbstractItcbPortletApplication.this);
	}

	@Override
	public void handleEventRequest(EventRequest request,
			EventResponse response, Window window) {		
	}

	@Override
	public void handleResourceRequest(ResourceRequest request,
			ResourceResponse response, Window window) {
		if(mainWindow.getContent() == null || !mainWindow.getContent().getComponentIterator().hasNext()) {
			if (preferencesURL == null) {
				PortletURL url = response.createActionURL();
				try {
					url.setPortletMode(PortletMode.EDIT);
					url.setWindowState(WindowState.MAXIMIZED);
					preferencesURL = url.toString();
				}
				catch (PortletModeException e) {
					log.warn(e.getMessage());
				}
				catch (WindowStateException e) {
					log.warn(e.getMessage());
				}
			}
			
			mainWindow.removeAllComponents();
			
			// check mode in request and add the select the pageMappingProcessor accordingly
			try {
				if(currentPageMappingProcessor.isAutomaticDisplay()) {
					currentPageMappingProcessor.displayDefaultPage();
				}
			} 
			catch (Exception e) {
				try  {
					currentPageMappingProcessor.handleException(e);
				}
				catch(Exception errorInExceptionManagementException) {
					errorInExceptionManagementException.printStackTrace();
				}
			}
		}
	}

}
