package net.sf.itcb.common.portlet.vaadin.page.impl;

import java.io.Serializable;
import java.util.LinkedList;
import java.util.Map;

import javax.portlet.PortletRequest;
import javax.portlet.PortletSession;
import javax.servlet.ServletRequest;

import net.sf.itcb.common.business.core.ItcbApplicationContextHolder;
import net.sf.itcb.common.portlet.exceptions.PortletItcbException;
import net.sf.itcb.common.portlet.exceptions.PortletItcbExceptionMappingErrors;
import net.sf.itcb.common.portlet.portal.PortalAdapterProvider;
import net.sf.itcb.common.portlet.vaadin.AbstractItcbPortletApplication;
import net.sf.itcb.common.portlet.vaadin.exception.ExceptionHandlerMapping;
import net.sf.itcb.common.portlet.vaadin.page.AbstractItcbPortletPage;
import net.sf.itcb.common.portlet.vaadin.page.AbstractItcbPortletController;
import net.sf.itcb.common.portlet.vaadin.page.PageMappingProcessor;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.context.MessageSource;
import org.springframework.context.i18n.LocaleContextHolder;
import org.springframework.scheduling.annotation.Async;

import com.vaadin.terminal.ParameterHandler;
import com.vaadin.terminal.Terminal.ErrorEvent;
import com.vaadin.terminal.URIHandler;
import com.vaadin.terminal.VariableOwner;
import com.vaadin.terminal.gwt.server.ChangeVariablesErrorEvent;
import com.vaadin.ui.Panel;

/**
 * ITCB implementation of PageMappingProcessor
 * 
 * @author Pierre Le Roux
 *
 */
public class PageMappingProcessorImpl implements PageMappingProcessor, Serializable {

	private static final long serialVersionUID = 1L;
	
	protected final Logger log = LoggerFactory.getLogger(getClass());
	
	protected Map<String, AbstractItcbPortletPage> mapping;
	
	protected String defaultPageKey;

	protected ExceptionHandlerMapping exceptionHandlerMapping;
	
	protected Panel panel;
	
	protected PortletRequest portletRequest;
	
	protected ServletRequest servletRequest;
	
	protected AbstractItcbPortletApplication application;
	
	protected AbstractItcbPortletPage currentPage;
	
	protected LinkedList<String> pageHistory;
	
	protected static PortalAdapterProvider portalAdapterProvider = ItcbApplicationContextHolder.getContext().getBean("itcbPortalAdapterProvider", PortalAdapterProvider.class);
	
	/**
	 * By default, the default page is displayed at portlet initialization<br/>
	 * If the portlet doesn't want to use this processor at portlet initialization,
	 * Spring context has to be set to false in Spring configuration
	 */
	protected Boolean isAutomaticDisplay = true;
	
	
	public PageMappingProcessorImpl() {
		pageHistory = new LinkedList<String>();
	}
	
	@Override
	public void setMapping(Map<String, AbstractItcbPortletPage> mapping) {
		this.mapping = mapping;
	}
	
	@Override
	public void setDefaultPageKey(String defaultPageKey) {
		this.defaultPageKey = defaultPageKey;
	}
	
	@Override
	public void setIsAutomaticDisplay(Boolean isAutomaticDisplay) {
		this.isAutomaticDisplay = isAutomaticDisplay;
	}
	
	@Override
	public void setExceptionHandlerMapping(
			ExceptionHandlerMapping exceptionHandlerMapping) {
		this.exceptionHandlerMapping = exceptionHandlerMapping;
	}
	
	
	// Application Specific params
	
	@Override
	public void setRequest(PortletRequest request) {
		this.portletRequest = request;
		this.servletRequest = portalAdapterProvider.getPortalAdapter(request).getServletRequest(request);
	}
	
	@SuppressWarnings({ "unchecked" })
	@Override
	public void setRequest(ServletRequest servletRequest) {
		//MockPortletRequest mockRequest = new MockPortletRequest();
		// TODO Mock the portletRequest in order to make it available in servlet mode as if the user were in a portal
		//for (Iterator<Map.Entry<String,String>> iterator = servletRequest.getParameterMap().entrySet().iterator(); iterator.hasNext();) {
		//	Map.Entry<String, String> entry = (Map.Entry<String, String>) iterator.next();
			//mockRequest.addParameter(entry.getKey().toString(), entry.getValue().toString());
		//}
		//this.request = mockRequest;
		this.servletRequest=servletRequest;
	}
	
	@Override
	public void setPanel(Panel panel) {
		this.panel = panel;
	}
	
	@Override
	public void setApplication(AbstractItcbPortletApplication application) {
		this.application = application;
	}

	@Override
	public AbstractItcbPortletApplication getApplication() {
		return application;
	}
	
	@Override
	public void displayPage(String refPage, boolean reload) {
		if(reload) {
			displayPage(refPage, ReloadOrder.TRUE);
		}
		else {
			displayPage(refPage, ReloadOrder.FALSE);
		}
	}
	
	@Override
	public void displayPage(String refPage, boolean reload, boolean closeApplicationOnExit) {
		application.setHasToCloseApplicationOnExit(closeApplicationOnExit);
		displayPage(refPage, reload);
	}
	
	@Override
	public void displayPage(String refPage, ReloadOrder reload) {
		try  {
			log.debug("Display {}", refPage);
			AbstractItcbPortletPage page = mapping.get(refPage);	
			if(page == null) {
				MessageSource itcbPortletMessageSource = ItcbApplicationContextHolder.getContext().getBean("common-portletResources", MessageSource.class);
				throw new PortletItcbException(PortletItcbExceptionMappingErrors.COMMON_PORTLET_MISSING_PAGE, itcbPortletMessageSource.getMessage("common.portlet.exception.missing.page", new Object[] {refPage}, LocaleContextHolder.getLocale()));
			}
			if(currentPage != null && currentPage != page) {
				currentPage.setVisible(false);
			}
			
			if (page instanceof AbstractItcbPortletPage.Updatable && page.getParent() != null
					&& (reload.equals(ReloadOrder.FALSE) || 
							((reload.equals(ReloadOrder.IF_MODIFIED) && !page.hasChangedSinceLastDisplay())))) {
				((AbstractItcbPortletPage.Updatable)page).update();
			}
			else {
				loadPage(page, reload);
			}
			
			
			// Store the previous page before changing the printed page only if we are in a page. A controller is not considered as a page that has to be in history
			
			if (! (page instanceof AbstractItcbPortletController) ) {
					page.setVisible(true);
					currentPage=page;
			
				if(pageHistory.size() > 0) {
					if(refPage.compareTo(pageHistory.getLast()) != 0) {
						pageHistory.add(refPage);
					}
				}
				else {
					pageHistory.add(refPage);
				}
			}
		
		}
		catch(RuntimeException re) {
			throw re;
		}
		catch(final Exception e) {
			MessageSource itcbPortletMessageSource = ItcbApplicationContextHolder.getContext().getBean("common-portletResources", MessageSource.class);
			throw new PortletItcbException(PortletItcbExceptionMappingErrors.COMMON_PORTLET_LOADING_PAGE, itcbPortletMessageSource.getMessage("common.portlet.exception.loading.page", new Object[] {refPage, e.getMessage()}, LocaleContextHolder.getLocale()));
		}
		
	}

	@Override
	public void displayPage(String refPage, ReloadOrder reload,
			boolean closeApplicationOnExit) {
		application.setHasToCloseApplicationOnExit(closeApplicationOnExit);
		displayPage(refPage, reload);
	}
	
	
	@Override
	public void displayDefaultPage() {
		log.debug("Display the default page {}", defaultPageKey);
		if(defaultPageKey != null) {
			panel.setContent(null);
			for (AbstractItcbPortletPage page : mapping.values()) {
				page.setParent(null);
			}
			displayPage(defaultPageKey, true);
		}
		else {
			if(log.isDebugEnabled()) {
				log.warn("No default page configured. PageMappingProcessor mecanism not used. Set the defaultPageKey property in your spring context.");
			}
		}
	}
	
	@Override
	public void displayPreviousPage() {
		if(pageHistory.size() >= 2) {
			String previousPage = pageHistory.get(pageHistory.size() - 2);
			
			displayPage(previousPage, false);
			// Displaying page succeed if we execute following lines
			pageHistory.pollLast(); // We delete the new record added for the previous page
			pageHistory.pollLast(); // We delete the current page we were on
			// the last page in list is the previous page which becomes the current page.
		}
		else {
			log.error("You can't display a previous page because no previous page exists");
		}
	}
	
	@Override
	public boolean hasPreviousPage() {
		if(pageHistory.size() >= 1) {
			return true;
		}
		return false;
	}
	
	@Override
	public void setSessionAttribute (String key, Object value) {
		portletRequest.getPortletSession().setAttribute(key, value);
	}
	
	@Override
	@SuppressWarnings("unchecked")
	public <T> T getSessionAttribute(String key, Class<T> requiredType) {
		return (T) portletRequest.getPortletSession().getAttribute(key );
	}
	
	@Override
	@SuppressWarnings("unchecked")
	public <T> T getSharedSessionAttribute(String key, Class<T> requiredType) {
		return (T) portletRequest.getPortletSession().getAttribute(key, PortletSession.APPLICATION_SCOPE);
	}
	
	@Override
	public PortletRequest getRequest() {
		return portletRequest;
	}

	@Override
	public boolean isAutomaticDisplay() {
		return isAutomaticDisplay;
	}
	
	@Override
	public void handleException(Throwable e) {
		if(e instanceof RuntimeException) {
			throw (RuntimeException)e;
		}
		else {
			
			MessageSource itcbPortletMessageSource = ItcbApplicationContextHolder.getContext().getBean("common-portletResources", MessageSource.class);
			throw new PortletItcbException(PortletItcbExceptionMappingErrors.COMMON_PORTLET_LOADING_PAGE, itcbPortletMessageSource.getMessage("common.portlet.exception.loading.page", new Object[] {e.getMessage()}, LocaleContextHolder.getLocale()));
		}
	}
	
	@Override
	public void handleError(ErrorEvent error) {
		try {
			if(exceptionHandlerMapping != null) {
				exceptionHandlerMapping.handleError(error, this);
			}
			else {
				Throwable t = error.getThrowable().getCause();
				if(t != null) {
					log.error(t.getMessage(), t);
				}
			}
		}
		catch(Exception exceptionInExceptionManagement) {
			log.error("The exception has not been handled well {}. This Exception occured {}", error.getThrowable(), exceptionInExceptionManagement);
		}
	}

	@Override
	public String getRequestParameter(String key) {
		String value = portletRequest.getParameter(key);
		if(value == null) {
			value=servletRequest.getParameter(key);
		}
		return value;
	}

	@Override
	public Object getErrorOwner(ErrorEvent event) {
		Object owner = null;
        if (event instanceof VariableOwner.ErrorEvent) {
            owner = ((VariableOwner.ErrorEvent) event).getVariableOwner();
        } else if (event instanceof URIHandler.ErrorEvent) {
            owner = ((URIHandler.ErrorEvent) event).getURIHandler();
        } else if (event instanceof ParameterHandler.ErrorEvent) {
            owner = ((ParameterHandler.ErrorEvent) event).getParameterHandler();
        } else if (event instanceof ChangeVariablesErrorEvent) {
            owner = ((ChangeVariablesErrorEvent) event).getComponent();
        }

        return owner;
		
	}

	@SuppressWarnings("unchecked")
	@Override
	public <T extends AbstractItcbPortletPage> T getPage(String refPage,
			Class<T> requiredType) {
		return (T) mapping.get(refPage);
	}

	@Override
	@Async
	/**
	 * This function is called with async method in order to construct the page serverside without time issue for the user
	 */
	public void preloadPage(String refPage, ReloadOrder reload,
			boolean closeApplicationOnExit) {
		try  {
			log.debug("Preload {}", refPage);
			AbstractItcbPortletPage page = mapping.get(refPage);	
			if(page == null) {
				MessageSource itcbPortletMessageSource = ItcbApplicationContextHolder.getContext().getBean("common-portletResources", MessageSource.class);
				throw new PortletItcbException(PortletItcbExceptionMappingErrors.COMMON_PORTLET_MISSING_PAGE, itcbPortletMessageSource.getMessage("common.portlet.exception.missing.page", new Object[] {refPage}, LocaleContextHolder.getLocale()));
			}
			loadPage(page, reload);		
		}
		catch(final Exception e) {
			log.error(e.getMessage());
		}
		
	}

	private void loadPage(AbstractItcbPortletPage page, ReloadOrder reload)
			throws Exception {
		if(page != null && 
				(reload == ReloadOrder.TRUE || 
						(reload == ReloadOrder.IF_MODIFIED && page.hasChangedSinceLastDisplay()))) {
			page.setPageMappingProcessor(this);
			page.removeAllComponents();
			page.initPage();
			if(page.getParent() == null) {
				panel.addComponent(page);
			}
		}
		if(page.getPageMappingProcessor() == null || page.getParent() == null || (!page.getComponentIterator().hasNext() && !(page instanceof AbstractItcbPortletController) ) ) {
			page.setPageMappingProcessor(this);
			page.initPage();
			panel.addComponent(page);
		}
	}	
	
	
	
}
